@@ -8,7 +8,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tkMacKeyboard.c,v 1.4 2000/02/09 02:13:53 hobbs Exp $
+ * RCS: @(#) $Id: tkMacKeyboard.c,v 1.6 2000/04/17 02:16:58 jingham Exp $
  */
 
 #include "tkInt.h"
@@ -72,7 +72,6 @@ static Ptr KCHRPtr;			/* Pointer to 'KCHR' resource. */
  * Prototypes for static functions used in this file.
  */
 static void	InitKeyMaps _ANSI_ARGS_((void));
-static void	InitKeymapInfo _ANSI_ARGS_((TkDisplay *dispPtr));
 
 
 /*
@@ -347,6 +346,8 @@ XStringToKeysym(
  *	The function XKeysymToKeycode is only used by tkTest.c and
  *	currently only implementes the support for keys used in the
  *	Tk test suite.
+ *      FIXME - This is no longer true.  This function is now used in
+ *      "event generate" so we really should make it work.
  *
  * Results:
  *	None.
@@ -466,7 +467,7 @@ TkpGetKeySym(dispPtr, eventPtr)
      */
 
     if (dispPtr->bindInfoStale) {
-	InitKeymapInfo(dispPtr);
+	TkpInitKeymapInfo(dispPtr);
     }
 
     /*
@@ -519,7 +520,7 @@ TkpGetKeySym(dispPtr, eventPtr)
 /*
  *--------------------------------------------------------------
  *
- * InitKeymapInfo --
+ * TkpInitKeymapInfo --
  *
  *	This procedure is invoked to scan keymap information
  *	to recompute stuff that's important for binding, such
@@ -535,8 +536,8 @@ TkpGetKeySym(dispPtr, eventPtr)
  *--------------------------------------------------------------
  */
 
-static void
-InitKeymapInfo(dispPtr)
+void
+TkpInitKeymapInfo(dispPtr)
     TkDisplay *dispPtr;		/* Display for which to recompute keymap
 				 * information. */
 {
